import os

import pytest

from kestrel.exceptions import KestrelSyntaxError
from kestrel.exceptions import VariableNotExist
from kestrel.session import Session


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


def test_unexpected_chars(fake_bundle_file):
    with Session(debug_mode=True) as session:
        with pytest.raises(KestrelSyntaxError) as einfo:
            session.execute(
                f"get ? network-traffic from file://{fake_bundle_file} where"
            )
        err = einfo.value
        assert err.line == 1
        assert err.column == 5
        assert err.expected == ["ENTITY_TYPE"]


def test_unexpected_token(fake_bundle_file):
    with Session(debug_mode=True) as session:
        with pytest.raises(KestrelSyntaxError) as einfo:
            session.execute(f"get from file://{fake_bundle_file} where")
        err = einfo.value
        assert err.line == 1
        assert err.column == 10
        assert set(err.expected) == set(["FROM", "WHERE"])


def test_unexpected_eol(fake_bundle_file):
    with Session(debug_mode=True) as session:
        with pytest.raises(KestrelSyntaxError) as einfo:
            session.execute(f"get process from file://{fake_bundle_file}")
        err = einfo.value
        assert err.line == 1
        assert err.column == 18
        assert err.expected == ["WHERE"]


def test_undefined_variable(fake_bundle_file):
    with Session(debug_mode=True) as session:
        with pytest.raises(VariableNotExist) as einfo:
            session.execute(f"get process from file://{fake_bundle_file} where pid=abc.pid")
        err = einfo.value
        assert err.var_name == 'abc'


def test_garbage():
    with Session(debug_mode=True) as session:
        with pytest.raises(KestrelSyntaxError) as einfo:
            session.execute("garbage")
        err = einfo.value
        assert err.expected == ["EQUAL"]



import json
import logging
import os
import pytest
import pathlib
import shutil
import tempfile
import kestrel
import kestrel_datasource_stixshifter
import pandas as pd

from kestrel.session import Session


def get_df(session, var_name):
    return pd.DataFrame.from_records(session.get_variable(var_name))


def execute(session, script):
    result = session.execute(script)
    if isinstance(result, str):
        assert not result.startswith("[ERROR]")


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


@pytest.fixture
def fake_bundle_2():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle_2.json")


@pytest.fixture
def fake_bundle_3():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle_3.json")


@pytest.fixture
def cbcloud_powershell_bundle():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "powershell_search_stix_result.json")


def test_session_1(fake_bundle_file):
    with Session(debug_mode=True) as session:
        execute(
            session,
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port < 10000]""",
        )
        conns = get_df(session, "conns")
        assert len(conns.index) == 100
        execute(session, "sort conns by dst_port asc")
        s = get_df(session, "_")
        assert len(s.index) == 100
        assert s.iloc[0]["dst_port"] == 22
        execute(session, "group conns by dst_port")
        s = get_df(session, "_")
        assert len(s.index) == 5
        port_3128 = s[(s["dst_port"] == 3128)]
        assert len(port_3128.index) == 1

        conns_sym = session.symtable["conns"]
        conns_dict = dict(conns_sym)
        assert conns_dict["type"] == conns_sym.type
        assert conns_dict["entity_table"] == conns_sym.entity_table


def test_session_timeframe(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session = Session()
        script = f"""conns = get network-traffic
                     from file://{fake_bundle_file}
                     where [network-traffic:dst_port = 22] START t'2020-06-30T19:25:00.000Z' STOP t'2020-06-30T19:26:00.000Z'"""
        execute(session, script)
        conns = get_df(session, "conns")
        assert len(conns.index) == 7


@pytest.mark.parametrize(
    "sco_type, prop, op, value, count",
    [
        ("ipv4-addr", "value", "=", "'192.168.121.121'", 1),
        ("network-traffic", "src_ref.value", "=", "'192.168.121.121'", 1),
        ("network-traffic", "dst_port", "=", 22, 29),
        ("user-account", "account_login", "=", "'henry'", 1),
        ("user-account", "account_login", "LIKE", "'hen%'", 1),
        ("user-account", "account_login", "=", "'zane'", 0),
    ],
)
def test_session_simple(fake_bundle_file, sco_type, prop, op, value, count):
    with Session(debug_mode=True) as session:
        script = f"""result = get {sco_type} from file://{fake_bundle_file} where [{sco_type}:{prop} {op} {value}]"""
        execute(session, script)
        result = get_df(session, "result")
        assert len(result.index) == count


@pytest.mark.parametrize(
    "sco_type, pattern, count",
    [
        (
            "network-traffic",
            "[network-traffic:dst_ref.value = '10.0.0.91' AND network-traffic:dst_port = 22]",
            3,
        ),
        (
            "network-traffic",
            "[network-traffic:dst_ref.value = '10.0.0.91' OR network-traffic:dst_port = 22]",
            35,
        ),
    ],
)
def test_session_complex(fake_bundle_file, sco_type, pattern, count):
    with Session(debug_mode=True) as session:
        script = f"""result = get {sco_type} from file://{fake_bundle_file} where {pattern}"""
        execute(session, script)
        result = get_df(session, "result")
        assert len(result.index) == count


def test_generated_pattern(fake_bundle_file, fake_bundle_2):
    with Session(debug_mode=True) as session:
        script = f"""conns_a = get network-traffic
             from file://{fake_bundle_file}
         where [network-traffic:dst_ref.value = '10.0.0.134']"""
        execute(session, script)
        conns_a = get_df(session, "conns_a")
        script = f"""conns_b = get network-traffic
             from file://{fake_bundle_2}
         where [network-traffic:dst_port = conns_a.dst_port]"""
        execute(session, script)
        conns_b = get_df(session, "conns_b")
        conns_b.to_csv("conns_b.csv")
        # time range failed it
        assert len(conns_b.index) == 0
        assert os.path.exists("conns_b.csv")
        os.remove("conns_b.csv")


def test_generated_pattern_match(fake_bundle_file, fake_bundle_3):
    with Session(debug_mode=True) as session:
        script = f"""conns_a = get network-traffic
             from file://{fake_bundle_file}
         where [network-traffic:dst_ref.value = '10.0.0.134']"""
        execute(session, script)
        conns_a = get_df(session, "conns_a")
        script = f"""conns_b = get network-traffic
             from file://{fake_bundle_3}
         where [network-traffic:dst_port = conns_a.dst_port]"""
        execute(session, script)
        conns_b = get_df(session, "conns_b")
        conns_b.to_csv("conns_b.csv")
        # time range not tested since it is only generated for udi data sources
        assert len(conns_b.index) == 3
        # assert len(conns_b.index) == 2  # 2/3 matches due to time range
        assert os.path.exists("conns_b.csv")
        os.remove("conns_b.csv")


def test_disp_column_order(fake_bundle_file, caplog):
    #caplog.set_level(logging.DEBUG)
    with Session(debug_mode=True) as session:
        execute(
            session,
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port < 10000]""",
        )
        # SCO type in attr names should be optional
        recs = session.execute(f"disp conns attr src_port, dst_port")[0]
        conns = recs.dataframe
        print(conns.head())
        cols = conns.columns.to_list()
        assert cols.index("src_port") < cols.index("dst_port")

def test_get_set_variable(fake_bundle_file):
    with Session() as session:
        # Create a normal var
        script = f"x = get ipv4-addr from file://{fake_bundle_file} where [ipv4-addr:value = '192.168.121.121']"
        execute(session, script)
        var_list = session.get_variable_names()
        assert "x" in var_list
        var_x = session.get_variable("x")
        assert len(var_x) == 1
        val = var_x[0]
        assert val["type"] == "ipv4-addr"
        assert val["value"] == "192.168.121.121"

        # Now create a new var using Session API
        names = ["alice", "bob", "carol"]
        session.create_variable("y", names, object_type="user-account")
        var_list = session.get_variable_names()
        assert "x" in var_list
        assert "y" in var_list
        var_y = session.get_variable("y")
        assert len(var_y) == 3
        print(var_y)
        val = var_y[0]
        assert val["type"] == "user-account"
        # Maybe this should be 'account_login'?
        assert (
            val["user_id"] in names
        )  # Order is not preserved, so it could be any of these


def test_session_runtime_dir():
    # standard session
    with Session(debug_mode=False) as session:
        runtime_directory = session.runtime_directory
        runtime_master_dirctory = session._get_runtime_directory_master()

        assert os.path.exists(session.runtime_directory)

        # if not executed in a clean env; it has previous debug dirs
        if runtime_master_dirctory.exists():
            d = pathlib.Path(session.runtime_directory).resolve()
            d_master = runtime_master_dirctory.resolve()
            assert d != d_master

    assert not os.path.exists(session.runtime_directory)

    # debug session
    with Session(debug_mode=True) as session:
        runtime_directory = session.runtime_directory
        runtime_master_dirctory = session._get_runtime_directory_master()

        assert os.path.exists(runtime_directory)

    assert os.path.exists(runtime_directory)

    if runtime_master_dirctory.exists():
        d = pathlib.Path(runtime_directory).resolve()
        d_master = runtime_master_dirctory.resolve()
        assert d == d_master

    # predefined runtime_dir session managed by session
    d = pathlib.Path(tempfile.gettempdir()) / "kestrel-runtime-test"
    d = d.resolve()
    if os.path.exists(d):
        shutil.rmtree(d)
    with Session(runtime_dir=d) as session:
        session = Session()
        assert os.path.exists(d)
    assert not os.path.exists(d)

    # predefined runtime_dir session not managed by session
    d = pathlib.Path(tempfile.gettempdir()) / "kestrel-runtime-test"
    d = d.resolve()
    pathlib.Path(d).mkdir(parents=True, exist_ok=True)
    with Session(runtime_dir=d) as session:
        assert os.path.exists(d)
    assert os.path.exists(d)


def test_session_debug_from_env():
    os.environ["KESTREL_DEBUG"] = "something"
    with Session() as session:
        assert session.debug_mode == True


def test_sha256_attr_name(cbcloud_powershell_bundle):
    # Make sure we can handle single quotes in attr names
    with Session() as session:
        script = (
            "x = get process"
            f" from file://{cbcloud_powershell_bundle}"
            " where [process:name = 'powershell.exe']"
        )
        execute(session, script)
        out = session.execute("DISP x ATTR binary_ref.hashes.'SHA-256'")
        df = out[0].dataframe
        assert (
            df["binary_ref.hashes.'SHA-256'"][0]
            == "de96a6e69944335375dc1ac238336066889d9ffc7d73628ef4fe1b1b160ab32c"
        )


def test_disp_after_group(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""
conns = get network-traffic from file://{fake_bundle_file}
    where [network-traffic:dst_port < 10000]
grouped = group conns by src_ref.value, dst_ref.value with count(src_ref.value) as count
"""
        )
        out = session.execute("DISP grouped ATTR src_ref.value, dst_ref.value, count")
        df = out[0].dataframe
        assert list(df.columns) == ["src_ref.value", "dst_ref.value", "count"]


def test_env_var_resolve(tmp_path):
    os.chdir(tmp_path)
    config_name = "abc.yaml"
    with open(config_name, "w") as config:
        config.write(r"""
language:
  default_variable: "_"
""")
    os.environ[kestrel.config.CONFIG_PATH_ENV_VAR] = config_name
    os.environ[kestrel_datasource_stixshifter.config.PROFILE_PATH_ENV_VAR] = config_name
    s = Session()
    full_path = os.path.join(os.getcwd(), config_name)
    assert os.environ[kestrel.config.CONFIG_PATH_ENV_VAR] == full_path 
    assert os.environ[kestrel_datasource_stixshifter.config.PROFILE_PATH_ENV_VAR] == full_path


def test_where_deref_network_traffic(fake_bundle_file):
    '''#290: expression does not work on attribute with reference'''
    with Session(debug_mode=True) as session:
        execute(
            session,
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port < 10000]""",
        )
        conns = get_df(session, "conns")
        assert len(conns.index) == 100
        execute(session, "c2 = conns where src_ref.value = '192.168.212.97'")
        c2 = session.get_variable("c2")
        assert len(c2) == 1
        assert c2[0]["dst_port"] == 22

        execute(session, "c2 = conns where src_ref.value = '192.168.121.121' and dst_ref.value = '10.0.0.134'")
        c2 = session.get_variable("c2")
        assert len(c2) == 1
        assert c2[0]["dst_port"] == 3128


def test_where_deref_process(cbcloud_powershell_bundle):
    '''#290: expression does not work on attribute with reference'''
    with Session() as session:
        script = (
            "x = get process"
            f" from file://{cbcloud_powershell_bundle}"
            " where [process:name = 'powershell.exe']"
        )
        execute(session, script)
        execute(session, "y = x where parent_ref.pid = 1544")
        y = session.get_variable("y")
        assert len(y) == 1
        assert y[0]["parent_ref.x_unique_id"] == "MYORGIDX-02629f16-00000608-00000000-1d71d10a09cc7c4"



"""A Kestrel session provides an isolated stateful runtime space for a huntflow.

A huntflow is the source code or script of a cyber threat hunt, which can be
developed offline in a text editor or interactively as the hunt goes. A Kestrel
session provides the runtime space for a huntflow that allows execution and
inspection of hunt statements in the huntflow. The :class:`Session` class in this
module supports both non-interactive and interactive execution of huntflows as
well as comprehensive APIs besides execution.

.. highlight:: python

Examples:
    A non-interactive execution of a huntflow::

        from kestrel.session import Session
        with Session() as session:
            open(huntflow_file) as hff:
                huntflow = hff.read()
            session.execute(huntflow)

    An interactive composition and execution of a huntflow::

        from kestrel.session import Session
        with Session() as session:
            try:
                hunt_statement = input(">>> ")
            except EOFError:
                print()
                break
            else:
                output = session.execute(hunt_statement)
                print(output)

    Export Kestrel variable to Python::

        from kestrel.session import Session
        huntflow = ""\"newvar = GET process
                               FROM stixshifter://workstationX
                               WHERE [process:name = 'cmd.exe']""\"
        with Session() as session:
            session.execute(huntflow)
            cmds = session.get_variable("newvar")
        for process in cmds:
            print(process["name"])

"""

import tempfile
import os
import getpass
import pathlib
import shutil
import uuid
import logging
import time
import math
import lark
import atexit
from contextlib import AbstractContextManager

from kestrel.exceptions import (
    KestrelSyntaxError,
    InvalidStixPattern,
    DebugCacheLinkOccupied,
)
from kestrel.syntax.parser import parse_kestrel
from kestrel.semantics.processor import semantics_processing
from kestrel.semantics.completor import do_complete
from kestrel.codegen import commands
from kestrel.codegen.display import DisplayBlockSummary
from kestrel.codegen.summary import gen_variable_summary
from kestrel.symboltable.symtable import SymbolTable
from kestrel.utils import (
    set_current_working_directory,
    resolve_path_in_kestrel_env_var,
    add_logging_handler,
)
from kestrel.config import load_config
from kestrel.datasource import DataSourceManager
from kestrel.analytics import AnalyticsManager
from firepit import get_storage
from firepit.exceptions import StixPatternError

_logger = logging.getLogger(__name__)


class Session(AbstractContextManager):
    """Kestrel Session class

    A session object needs to be instantiated to create a Kestrel runtime space.
    This is the foundation of multi-user dynamic composition and execution of
    huntflows. A Kestrel session has two important properties:

    - Stateful: a session keeps track of states/effects of statements that have
      been previously executed in this session, e.g., the values of previous
      established Kestrel variables. A session can invoke more than one
      :meth:`execute`, and each :meth:`execute` can process a block of Kestrel code,
      i.e., multiple Kestrel statements.

    - Isolated: each session is established in an isolated space (memory and
      file system):

      - Memory isolation is accomplished by OS process and memory space
        management automatically -- different Kestrel session instances will not
        overlap in memory.

      - File system isolation is accomplished with the setup and management of
        a temporary runtime directory for each session.

    Args:

      runtime_dir (str): to be used for :attr:`runtime_directory`.

      store_path (str): the file path or URL to initialize :attr:`store`.

      debug_mode (bool): to be assign to :attr:`debug_mode`.

    Attributes:

        session_id (str): The Kestrel session ID, which will be created as a random
          UUID if not given in the constructor.

        runtime_directory (str): The runtime directory stores session related
          data in the file system such as local cache of queried results,
          session log, and may be the internal store. The session will use
          a temporary directory derived from :attr:`session_id` if the path is
          not specified in constructor parameters.

        store (firepit.SqlStorage): The internal store used
          by the session to normalize queried results, implement cache, and
          realize the low level code generation. The store from the
          ``firepit`` package provides an operation abstraction
          over the raw internal database: either a local store, e.g., SQLite,
          or a remote one, e.g., PostgreSQL. If not specified from the
          constructor parameter, the session will use the default SQLite
          store in the :attr:`runtime_directory`.

        debug_mode (bool): The debug flag set by the session constructor. If
          True, a fixed debug link ``/tmp/kestrel`` of :attr:`runtime_directory`
          will be created, and :attr:`runtime_directory` will not be removed by
          the session when terminating.

        runtime_directory_is_owned_by_upper_layer (bool): The flag to specify
          who owns and manages :attr:`runtime_directory`. False by default,
          where the Kestrel session will manage session file system isolation --
          create and destory :attr:`runtime_directory`. If True, the runtime
          directory is created, passed in to the session constructor, and will
          be destroyed by the calling site.

        symtable (dict): The continuously updated *symbol table* of the running
          session, which is a dictionary mapping from Kestrel variable names
          ``str`` to their associated Kestrel internal data structure
          ``VarStruct``.

        data_source_manager (kestrel.datasource.DataSourceManager): The
          data source manager handles queries to all data source interfaces such as
          local file stix bundle and stix-shifter. It also stores previous
          queried data sources for the session, which is used for a syntax
          sugar when there is no data source in a Kestrel ``GET`` statement -- the
          last data source is implicitly used.

        analytics_manager (kestrel.analytics.AnalyticsManager): The analytics
          manager handles all analytics related operations such as executing an
          analytics or getting the list of analytics for code auto-completion.

    """

    def __init__(
        self, session_id=None, runtime_dir=None, store_path=None, debug_mode=False
    ):
        _logger.debug(
            f"Establish session with session_id: {session_id}, runtime_dir: {runtime_dir}, store_path:{store_path}, debug_mode:{debug_mode}"
        )

        resolve_path_in_kestrel_env_var()

        self.config = load_config()

        if session_id:
            self.session_id = session_id
        else:
            self.session_id = str(uuid.uuid4())

        self.debug_mode = (
            True
            if debug_mode or os.getenv(self.config["debug"]["env_var"], False)
            else False
        )

        # default value of runtime_directory ownership
        self.runtime_directory_is_owned_by_upper_layer = False

        # runtime (temporary) directory to store session-related data
        sys_tmp_dir = pathlib.Path(tempfile.gettempdir())
        if runtime_dir:
            if os.path.exists(runtime_dir):
                self.runtime_directory_is_owned_by_upper_layer = True
            else:
                pathlib.Path(runtime_dir).mkdir(parents=True, exist_ok=True)
            self.runtime_directory = runtime_dir
        else:
            tmp_dir = sys_tmp_dir / (
                self.config["session"]["cache_directory_prefix"]
                + str(os.getuid())
                + "-"
                + self.session_id
            )
            self.runtime_directory = tmp_dir.expanduser().resolve()
            if tmp_dir.exists():
                if tmp_dir.is_dir():
                    _logger.debug(
                        "Kestrel session with runtime_directory exists, reuse it."
                    )
                else:
                    _logger.debug(
                        "strange tmp file that uses kestrel session dir name, remove it."
                    )
                    os.remove(self.runtime_directory)
            else:
                _logger.debug(
                    f"create new session runtime_directory: {self.runtime_directory}."
                )
                tmp_dir.mkdir(parents=True, exist_ok=True)

        if self.debug_mode:
            self._setup_runtime_directory_master()

        self._logging_setup()

        # local database of SQLite or PostgreSQL
        if not store_path:
            # use the default local database in config.py
            local_database_path = self.config["session"]["local_database_path"]
            if "://" in local_database_path:
                store_path = local_database_path
            else:
                store_path = os.path.join(self.runtime_directory, local_database_path)
        self.store = get_storage(store_path, self.session_id)

        # Symbol Table
        # linking variables in syntax with internal data structure
        # handling fallback_var for the most recently accessed var
        # {"var": VarStruct}
        self.symtable = SymbolTable()

        self.data_source_manager = DataSourceManager(self.config)
        self.analytics_manager = AnalyticsManager(self.config)

        atexit.register(self.close)

    def execute(self, codeblock):
        """Execute a Kestrel code block.

        A Kestrel statement or multiple consecutive statements constitute a code
        block, which can be executed by this method. New Kestrel variables can be
        created in a code block such as ``newvar = GET ...``. Two types of Kestrel
        variables can be legally referred in a Kestrel statement in the code block:

        * A Kestrel variable created in the same code block prior to the reference.

        * A Kestrel variable created in code blocks previously executed by the
          session. The session maintains the :attr:`symtable` to keep the state
          of all previously executed Kestrel statements and their established Kestrel
          variables.

        Args:
            codeblock (str): the code block to be executed.

        Returns:
            A list of outputs that each of them is the output for each
            statement in the inputted code block.
        """
        ast = self.parse(codeblock)
        return self._execute_ast(ast)

    def parse(self, codeblock):
        """Parse a Kestrel code block.

        Parse one or multiple consecutive Kestrel statements (a Kestrel code block)
        into the abstract syntax tree. This could be useful for frontends that
        need to parse a statement *without* executing it in order to render
        some type of interface.

        Args:
            codeblock (str): the code block to be parsed.

        Returns:
            A list of dictionaries that each of them is an *abstract syntax
            tree* for one Kestrel statement in the inputted code block.
        """
        try:
            ast = parse_kestrel(
                codeblock,
                self.config["language"]["default_variable"],
                self.config["language"]["default_sort_order"],
            )
        except lark.UnexpectedEOF as err:
            # use `err.expected` as used in lark.exceptions.UnexpectedEOF
            raise KestrelSyntaxError(
                err.line, err.column, "end of line", "", err.expected
            )
        except lark.UnexpectedCharacters as err:
            # use `err.allowed` as used in lark.exceptions.UnexpectedCharacters
            raise KestrelSyntaxError(
                err.line, err.column, "character", err.char, err.allowed
            )
        except lark.UnexpectedToken as err:
            # use `err.accepts or err.expected` as used in lark.exceptions.UnexpectedToken
            raise KestrelSyntaxError(
                err.line, err.column, "token", err.token, err.accepts or err.expected
            )
        return ast

    def get_variable_names(self):
        """Get the list of Kestrel variable names created in this session."""
        return list(self.symtable.keys())

    def get_variable(self, var_name, deref=True):
        """Get the data of Kestrel variable ``var_name``, which is list of homogeneous entities (STIX SCOs)."""
        # In the future, consider returning a generator here?
        return self.symtable[var_name].get_entities(deref)

    def create_variable(self, var_name, objects, object_type=None):
        """Create a new Kestrel variable ``var_name`` with data in ``objects``.

        This is the API equivalent to Kestrel command ``NEW``, while allowing more
        flexible objects types (Python objects) than the objects serialized
        into text/JSON in the command ``NEW``.

        Args:

            var_name (str): The Kestrel variable to be created.

            objects (list): List of Python objects, currently support either a
              list of ``str`` or a list of ``dict``.

            object_type (str): The Kestrel entity type for the created Kestrel
              variable. It overrides the ``type`` field in ``objects``. If
              there is no ``type`` field in ``objects``, e.g., ``objects`` is a
              list of ``str``, this parameter is required.

        """
        virtual_stmt_ast = [
            {"command": "new", "output": var_name, "data": objects, "type": object_type}
        ]
        self._execute_ast(virtual_stmt_ast)

    def do_complete(self, code, cursor_pos):
        """Kestrel code auto-completion.

        Args:
            code (str): Kestrel code.
            cursor_pos (int): the position to start completion (index in ``code``).

        Returns:
            A list of suggested strings to complete the code.
        """
        return do_complete(
            code,
            cursor_pos,
            self.data_source_manager,
            self.analytics_manager,
            self.symtable,
        )

    def close(self):
        """Explicitly close the session.

        This may be executed by a context manager or when the program exits.
        """
        # Note there are two conditions that trigger this function, so it is probably executed twice
        # Be careful to write the logic in this function to avoid deleting nonexist files/dirs

        # remove logging handler
        if self.logging_handler:
            logging.getLogger().removeHandler(self.logging_handler)
            self.logging_handler.close()
            # ensure this does not executed twice
            self.logging_handler = None

        # close store/database
        if self.store:
            # release resources
            self.store.close()
            # ensure this does not executed twice
            self.store = None

        # manage temp folder for debug
        # ensure this does not execute if already executed
        if os.path.exists(self.runtime_directory):
            if not self.runtime_directory_is_owned_by_upper_layer:
                if self.debug_mode:
                    self._leave_exit_marker()
                    self._remove_obsolete_debug_folders()
                else:
                    shutil.rmtree(self.runtime_directory)

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def _execute_ast(self, ast):
        displays = []
        new_vars = []

        start_exec_ts = time.time()
        for stmt in ast:
            try:
                # semantic checking and unfolding
                semantics_processing(
                    stmt,
                    self.symtable,
                    self.store,
                    self.data_source_manager,
                    self.config,
                )

                # code generation and execution
                execute_cmd = getattr(commands, stmt["command"])

                # set current working directory for each command execution
                # use this to implicitly pass runtime_dir as an argument to each command
                # the context manager switch back cwd when the command execution completes
                with set_current_working_directory(self.runtime_directory):
                    output_var_struct, display = execute_cmd(stmt, self)

            # exception completion
            except StixPatternError as e:
                raise InvalidStixPattern(e.stix) from e

            # post-processing: symbol table update
            if output_var_struct is not None:
                output_var_name = stmt["output"]
                self._update_symbol_table(output_var_name, output_var_struct)

                if output_var_name != self.config["language"]["default_variable"]:
                    if output_var_name in new_vars:
                        new_vars.remove(output_var_name)
                    new_vars.append(output_var_name)

            if display is not None:
                displays.append(display)

        end_exec_ts = time.time()
        execution_time_sec = math.ceil(end_exec_ts - start_exec_ts)

        if self.config["session"]["show_execution_summary"] and new_vars:
            vars_summary = [
                gen_variable_summary(vname, self.symtable[vname]) for vname in new_vars
            ]
            displays.append(DisplayBlockSummary(vars_summary, execution_time_sec))

        return displays

    def _update_symbol_table(self, output_var_name, output_var_struct):
        self.symtable[output_var_name] = output_var_struct
        self.symtable[self.config["language"]["default_variable"]] = output_var_struct

    def _leave_exit_marker(self):
        runtime_dir = pathlib.Path(self.runtime_directory)
        exit_marker = runtime_dir / self.config["debug"]["session_exit_marker"]
        exit_marker.touch()

    def _remove_obsolete_debug_folders(self):
        # will only clean debug cache directories under system temp directory

        # [(cache_dir, timestamp)]
        exited_sessions = []

        for x in pathlib.Path(tempfile.gettempdir()).iterdir():
            if x.is_dir() and x.parts[-1].startswith(
                self.config["session"]["cache_directory_prefix"]
                + str(os.getuid())
                + "-"
            ):
                marker = x / self.config["debug"]["session_exit_marker"]
                if marker.exists():
                    exited_sessions.append((x, marker.stat().st_mtime))

        # preserve the newest self.config["debug"]["maximum_exited_session"] debug sessions
        exited_sessions.sort(key=lambda x: x[1])
        for x, _ in exited_sessions[: -self.config["debug"]["maximum_exited_session"]]:
            shutil.rmtree(x)

    def _get_runtime_directory_master(self):
        sys_tmp_dir = pathlib.Path(tempfile.gettempdir())

        user_suffix = None
        for f in [getpass.getuser, os.getuid]:
            if not user_suffix:
                try:
                    user_suffix = f()
                except:
                    pass
        if not user_suffix:
            user_suffix = "noUID"

        return sys_tmp_dir / (
            self.config["debug"]["cache_directory_prefix"] + user_suffix
        )

    def _setup_runtime_directory_master(self):
        master_dir = self._get_runtime_directory_master()

        # master_dir.exists() should not be used
        # it will return False for broken link
        try:
            master_dir.unlink()
        except FileNotFoundError:
            pass
        except PermissionError:
            raise DebugCacheLinkOccupied(master_dir.resolve())

        master_dir.symlink_to(self.runtime_directory)

    def _logging_setup(self):
        log_file_name = self.config["session"]["log_path"]
        log_file_path = os.path.join(self.runtime_directory, log_file_name)
        handler = logging.FileHandler(log_file_path)
        self.logging_handler = add_logging_handler(handler, self.debug_mode)



import json
import logging
import os
import pytest
import pathlib
import shutil
import tempfile
import pandas as pd

from kestrel.session import Session


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


def test_group_srcref(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute("src_grps = group conns by src_ref.value")
        assert "src_grps" in session.get_variable_names()
        src_grps = session.get_variable("src_grps")
        assert src_grps is not None


def test_group_src_dst(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute(
            (
                "grps = group conns by src_ref.value, dst_ref.value"
            )
        )
        assert "grps" in session.get_variable_names()
        grps = session.get_variable("grps")
        assert grps is not None
        assert len(grps) == 94  # 94 unique src-dst pairs in test bundle


@pytest.mark.parametrize(
    "agg_func, attr, expected",
    [
        ("max", "dst_ref.value", "max_dst_ref.value"),
        ("count", "dst_ref.value", "count_dst_ref.value"),
        ("nunique", "dst_ref.value", "nunique_dst_ref.value"),
    ],
)
def test_group_srcref_agg(fake_bundle_file, agg_func, attr, expected):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute(
            (
                "src_grps = group conns by src_ref.value"
                f" with {agg_func}({attr})"
            )
        )
        assert "src_grps" in session.get_variable_names()
        src_grps = session.get_variable("src_grps")
        assert src_grps is not None
        assert expected in src_grps[0]


@pytest.mark.parametrize(
    "agg_func, attr, alias",
    [
        ("max", "dst_ref.value", "rand_value"),
        ("count", "dst_ref.value", "whatever"),
        ("nunique", "dst_ref.value", "unique_dests"),
    ],
)
def test_group_srcref_agg_alias(fake_bundle_file, agg_func, attr, alias):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute(
            (
                "src_grps = group conns by src_ref.value"
                f" with {agg_func}({attr}) as {alias}"
            )
        )
        assert "src_grps" in session.get_variable_names()
        src_grps = session.get_variable("src_grps")
        assert src_grps is not None
        assert alias in src_grps[0]


def test_group_nt_binned_timestamp(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute("ts_conns = TIMESTAMPED(conns)")
        session.execute(("conns_per_min = group ts_conns by bin(first_observed, 1m)"
                         " with count(dst_port) as count"))
        hist = session.get_variable("conns_per_min")
        assert len(hist) == 5
        assert hist[0]["first_observed_bin"] == "2020-06-30T19:25:00Z"
        assert hist[0]["count"] == 20
        assert hist[4]["first_observed_bin"] == "2020-06-30T19:29:00Z"
        assert hist[4]["count"] == 16


def test_group_nt_binned_port(fake_bundle_file):
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = get network-traffic
            from file://{fake_bundle_file}
            where [network-traffic:dst_port > 0]""",
        )

        session.execute(("port_hist = group conns by bin(src_port, 10000)"
                         " with count(src_port) as count"))
        hist = session.get_variable("port_hist")
        print(json.dumps(hist, indent=4))
        assert len(hist) == 3
        assert hist[0]["src_port_bin"] == 40000
        assert hist[0]["count"] == 4
        assert hist[1]["src_port_bin"] == 50000
        assert hist[1]["count"] == 69
        assert hist[2]["src_port_bin"] == 60000
        assert hist[2]["count"] == 27



# Python 3.10 supports this by default
# https://stackoverflow.com/questions/36286894/name-not-defined-in-type-annotation
from __future__ import annotations

from typeguard import typechecked

from typing import Tuple, Optional
import datetime
from abc import ABC, abstractmethod
from firepit.query import Column, Filter, Predicate, Query
from firepit.sqlstorage import get_path_joins
from firepit.timestamp import timefmt
from kestrel.syntax.utils import merge_timeranges
from kestrel.syntax.reference import (
    Reference,
    deref_and_flatten_value_to_list,
    value_to_stix,
)
from kestrel.exceptions import (
    KestrelNotImplemented,
    InvalidECGPattern,
    KestrelInternalError,
)


def _merge_queries(lhs: Query, op: str, rhs: Query):
    result = Query(lhs.table)
    result.joins = lhs.joins + rhs.joins
    lpred = lhs.where[0].preds[0]
    rpred = rhs.where[0].preds[0]
    result.where.append(Filter([Predicate(lpred, op, rpred)]))
    return result


class ExtCenteredGraphConstruct(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def to_stix(self, center_entity_type: str):
        pass

    @abstractmethod
    def to_firepit(self, center_entity_type: str):
        pass

    @abstractmethod
    def deref(self, deref_func):
        pass


@typechecked
class ExtCenteredGraphPattern(ExtCenteredGraphConstruct):
    def __init__(self, graph: ExtCenteredGraphConstruct):
        self.graph = graph
        self.timerange = None
        self.center_entity_type = None

    def __str__(self):
        return (
            f"graph: {self.graph}, "
            f"timerange: {self.timerange}, "
            f"center_entity_type: {self.center_entity_type}"
        )

    def add_center_entity(self, center_entity_type: str):
        self.center_entity_type = center_entity_type

    def prune_away_centered_graph(self, center_entity_type):
        # only leave the disconnected extended graph components
        if self.graph is not None:
            if_preserve_func = _make_extract_func(center_entity_type, "ext")
            self.graph = self.graph.prune(if_preserve_func)

    def prune_away_extended_graph(self, center_entity_type):
        # only leave the connected centered graph components
        if self.graph is not None:
            if_preserve_func = _make_extract_func(center_entity_type, "center")
            self.graph = self.graph.prune(if_preserve_func)

    def to_stix(
        self,
        timerange: Optional[Tuple[datetime.datetime, datetime.datetime]],
        timeadj: Optional[Tuple[datetime.timedelta, datetime.timedelta]],
    ):
        # timerange: user-specified timerange from Kestrel command
        # timeadj: time adjustments for START and STOP

        # before calling this, make sure:
        # 1. called deref()
        # 2. called add_center_entity()

        if self.center_entity_type is None:
            raise KestrelInternalError(
                "should run add_center_entity() before to_stix()"
            )

        if self.graph is None:
            inner = ""
        else:
            inner = self.graph.to_stix(self.center_entity_type)
        body = "[" + inner + "]"

        if timerange:
            tr = timerange
        elif self.timerange:
            tr = self.timerange
            if timeadj:
                tr = (tr[0] + timeadj[0], tr[1] + timeadj[1])
        else:
            tr = None

        if tr:
            tr_stix = f" START t'{timefmt(tr[0])}' STOP t'{timefmt(tr[1])}'"
        else:
            tr_stix = ""

        return body + tr_stix

    def to_firepit(self):
        if self.center_entity_type is None:
            raise KestrelInternalError(
                "should run add_center_entity() before to_firepit()"
            )
        if self.graph is None:
            return None
        else:
            return self.graph.to_firepit(self.center_entity_type)

    def deref(self, deref_func, get_timerange_func):
        if self.graph is not None:
            self.timerange = self.graph.deref(deref_func, get_timerange_func)

    def extend(self, junction_type: str, other_ecgp: Optional[ExtCenteredGraphPattern]):
        if other_ecgp is not None and other_ecgp.graph is not None:
            if self.center_entity_type is None:
                self.center_entity_type = other_ecgp.center_entity_type
            elif other_ecgp.center_entity_type is None:
                pass
            elif self.center_entity_type != other_ecgp.center_entity_type:
                raise InvalidECGPattern(
                    "could not merge ECGPs with different center entities:"
                    + self.center_entity_type
                    + ", "
                    + other_ecgp.center_entity_type
                )

            self.timerange = merge_timeranges((self.timerange, other_ecgp.timerange))

            junction_type = junction_type.upper()

            if junction_type == "AND":
                self.graph = ECGPJunction("AND", self.graph, other_ecgp.graph)
            elif junction_type == "OR":
                self.graph = ECGPJunction("OR", self.graph, other_ecgp.graph)
            else:
                raise KestrelInternalError(
                    f'Junction type {junction_type} not supported besides "AND", "OR".'
                )


@typechecked
class ECGPJunction(ExtCenteredGraphConstruct):
    def __init__(
        self,
        relation: str,
        lhs: ExtCenteredGraphConstruct,
        rhs: ExtCenteredGraphConstruct,
    ):
        # relation: "AND" | "OR"
        self.lhs = lhs
        self.rhs = rhs
        relation = relation.upper()
        if relation == "AND":
            self.relation = "AND"
        elif relation == "OR":
            self.relation = "OR"
        else:
            raise KestrelInternalError("unsupported relation for ECGPJunction()")

    def __str__(self):
        return f"({self.lhs}) {self.relation} ({self.rhs})"

    def prune(self, if_preserve_func):
        self.lhs = self.lhs.prune(if_preserve_func)
        self.rhs = self.rhs.prune(if_preserve_func)
        if self.lhs is None:
            return self.rhs
        elif self.rhs is None:
            return self.lhs
        else:
            return self

    def to_stix(self, center_entity_type: str):
        return (
            "("
            + self.lhs.to_stix(center_entity_type)
            + " "
            + self.relation
            + " "
            + self.rhs.to_stix(center_entity_type)
            + ")"
        )

    def to_firepit(self, center_entity_type: str):
        return _merge_queries(
            self.lhs.to_firepit(center_entity_type),
            self.relation,
            self.rhs.to_firepit(center_entity_type),
        )

    def deref(self, deref_func, get_timerange_func):
        ltr = self.lhs.deref(deref_func, get_timerange_func)
        rtr = self.rhs.deref(deref_func, get_timerange_func)
        return merge_timeranges((ltr, rtr))


@typechecked
class ECGPComparison(ExtCenteredGraphConstruct):
    def __init__(self, entity_attribute: str, operator: str, value, entity_type=None):
        self.etype = entity_type
        self.attribute = entity_attribute
        self.op = operator.upper()
        self.value = value
        if isinstance(self.value, list):
            if self.op not in ("IN", "NOT IN"):
                raise InvalidECGPattern(
                    'a list should be paired with the operator "IN"'
                )
        elif not isinstance(self.value, Reference):
            if self.op in ("IN", "NOT IN"):
                raise InvalidECGPattern(
                    'inappropriately pair operator "IN" with literal'
                )

    def __str__(self):
        return f"{self.etype}:{self.attribute} {self.op} {self.value}"

    def prune(self, if_preserve_func):
        if if_preserve_func(self.etype):
            return self
        else:
            return None

    def to_stix(self, center_entity_type: str):
        if not self.etype:
            self.etype = center_entity_type
        return f"{self.etype}:{self.attribute} {self.op} {value_to_stix(self.value)}"

    def to_firepit(self, center_entity_type: str):
        if self.etype and self.etype != center_entity_type:
            raise KestrelNotImplemented("firepit filtering on linked entities")
        joins, target_type, target_attr = get_path_joins(
            None, self.etype, self.attribute
        )
        if target_type:
            attribute = Column(target_attr, table=target_type)
        else:
            attribute = self.attribute
        qry = Query()
        qry.joins = joins
        qry.where = [Filter([Predicate(attribute, self.op, self.value)])]
        return qry

    def deref(self, deref_func, get_timerange_func):
        xs, tr = deref_and_flatten_value_to_list(
            self.value, deref_func, get_timerange_func
        )
        if len(xs) == 0:
            raise InvalidECGPattern("empty value after deref of {self.value}")
        elif len(xs) == 1:
            self.value = xs[0]
            if self.op == "IN":
                self.op = "="
            elif self.op == "NOT IN":
                self.op = "!="
        else:
            self.value = xs
            if self.op in ("=", "==", "IN"):
                self.op = "IN"
            elif self.op in ("!=", "NOT IN"):
                self.op = "NOT IN"
            else:
                raise InvalidECGPattern(
                    "operator {self.op} incompatible with value {self.value}"
                )
        return tr


def _make_extract_func(center_entity_type, preserve_center_or_ext: str):
    def if_preserve_func(entity_type):
        if entity_type is None:
            connected_to_center = True
        elif entity_type == center_entity_type:
            connected_to_center = True
        else:
            connected_to_center = False

        if preserve_center_or_ext == "center":
            return connected_to_center
        elif preserve_center_or_ext == "ext":
            return not connected_to_center
        else:
            raise KestrelInternalError(
                "unsupported argument in _make_extract_func: {prune_center_or_ext}"
            )

    return if_preserve_func



import pytest

from kestrel.session import Session
from kestrel.exceptions import MissingEntityType


def test_new_with_full_json():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "explorer.exe"]
        if v[0]["name"] == "cmd.exe":
            assert v[0]["pid"] == "123"
        else:
            assert v[0]["pid"] == "99"


def test_new_with_json_no_type():
    with Session() as s:
        stmt = """
newvar = NEW process [ {"name": "cmd.exe", "pid": "123"}
                     , {"name": "explorer.exe", "pid": "99"}
                     ]
"""
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "explorer.exe"]
        if v[0]["name"] == "cmd.exe":
            assert v[0]["pid"] == "123"
        else:
            assert v[0]["pid"] == "99"


def test_new_with_json_no_type_to_fail():
    with Session() as s:
        stmt = """
newvar = NEW [ {"name": "cmd.exe", "pid": "123"}
             , {"name": "explorer.exe", "pid": "99"}
             ]
"""
        with pytest.raises(MissingEntityType) as e:
            s.execute(stmt)


def test_new_with_list_of_strings():
    with Session() as s:
        stmt = (
            """newvar = NEW process ["cmd.exe", "explorer.exe", "google-chrome.exe"]"""
        )
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 3
        assert v[0]["type"] == "process"
        assert sorted([i["name"] for i in v]) == [
            "cmd.exe",
            "explorer.exe",
            "google-chrome.exe",
        ]


def test_new_list_of_strings_without_type_to_fail():
    with Session() as s:
        stmt = """newvar = NEW ["cmd.exe", "explorer.exe", "google-chrome.exe"]"""
        with pytest.raises(MissingEntityType) as e:
            s.execute(stmt)


def test_new_with_int_pid():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": 123}
             , {"type": "process", "name": "explorer.exe", "pid": 99}
             ]
"""
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "explorer.exe"]
        if v[0]["name"] == "cmd.exe":
            assert v[0]["pid"] == 123
        else:
            assert v[0]["pid"] == 99


def test_new_with_missing_field():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe"}
             ]
"""
        s.execute(stmt)
        v = sorted(s.get_variable("newvar"), key=lambda d: d["name"])
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "explorer.exe"]
        assert v[0]["pid"] == "123"
        assert v[1]["pid"] == None


def test_new_with_missing_field_first():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        s.execute(stmt)
        v = sorted(s.get_variable("newvar"), key=lambda d: d["name"])
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["name"] in ["cmd.exe", "explorer.exe"]
        assert v[0]["pid"] == None
        assert v[1]["pid"] == "99"



import pytest
import os

from kestrel.session import Session


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "doctored-1k.json")


def test_save_parquet_gz(tmp_path):
    save_path = tmp_path / "test_save_data.parquet.gz"
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_procs.parquet.gz"
    )
    stmt_save = f"newvar = LOAD {data_file_path} SAVE newvar TO {save_path}"
    stmt_load = f"newload = LOAD {save_path}"

    with Session() as s:
        s.execute(stmt_save)
    assert save_path.exists()

    with Session() as s:
        s.execute(stmt_load)
        v = s.get_variable("newload")
        assert len(v) == 5
        assert v[0]["type"] == "process"
        assert v[0]["name"] == "reg.exe"


def test_save_network_traffic_v4(tmp_path, fake_bundle_file):
    save_path = tmp_path / "conns.csv"
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = GET network-traffic
                        FROM file://{fake_bundle_file}
                        WHERE [network-traffic:dst_port > 0]""",
        )
        session.execute(f"SAVE conns TO {save_path}")
    assert save_path.exists()


def test_save_network_traffic_v4_v6(tmp_path, proc_bundle_file):
    save_path = tmp_path / "conns.csv"
    with Session(debug_mode=True) as session:
        session.execute(
            f"""conns = GET network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE [network-traffic:dst_port > 0]""",
        )
        session.execute(f"SAVE conns TO {save_path}")
    assert save_path.exists()



import pytest

from kestrel.session import Session
from kestrel.codegen import commands


def test_prefetch_config():
    with Session() as s:
        # by default, no excluded entity type in config
        assert commands._is_prefetch_allowed_in_config(s.config["prefetch"], "get", "process") == True
        # by default, no excluded entity type in config
        assert commands._is_prefetch_allowed_in_config(s.config["prefetch"], "find", "file") == True



import json
import os
import pytest

from kestrel.session import Session


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "doctored-1k.json")


def test_return_table_not_exist(fake_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        from file://{fake_bundle_file}
                        where [network-traffic:dst_port = 22]
                procs = FIND process CREATED conns
                """
        summaries = s.execute(stmt)
    correct_dict = {
        "display": "execution summary",
        "data": {
            "variables updated": [
                {
                    "VARIABLE": "conns",
                    "TYPE": "network-traffic",
                    "#(ENTITIES)": 29,
                    "#(RECORDS)": 29,
                    "ipv4-addr*": 58,
                    "network-traffic*": 0,
                    "user-account*": 29,
                },
                {
                    "VARIABLE": "procs",
                    "TYPE": "process",
                    "#(ENTITIES)": 0,
                    "#(RECORDS)": 0,
                    "ipv4-addr*": 0,
                    "network-traffic*": 0,
                    "user-account*": 0,
                },
            ],
            "footnotes": ["*Number of related records cached."],
        },
    }
    output_dict = summaries[0].to_dict()
    del output_dict["data"]["execution time"]
    assert output_dict == correct_dict


def test_find_srcs(fake_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        from file://{fake_bundle_file}
                        where dst_port = 22
                srcs = FIND ipv4-addr CREATED conns
                """
        s.execute(stmt)
        srcs = s.get_variable("srcs")
        assert len(srcs) == 24


def test_find_file_linked_to_process(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                files = FIND file LINKED procs
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 7 * 2
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 2 * 2


def test_find_file_linked_to_process_2():
    stixshifter_data_url = "https://raw.githubusercontent.com/opencybersecurityalliance/stix-shifter/develop/data/cybox"
    bundle = f"{stixshifter_data_url}/carbon_black/cb_observed_156.json"
    with Session() as s:
        stmt = f"""
                procs = get process
                        from {bundle}
                        where [process:name = 'svctest.exe']
                        files = FIND file LINKED procs
                """
        s.execute(stmt)
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 3


def test_find_file_loaded_by_process(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                files = FIND file LOADED BY procs
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 7 * 2
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 1


def test_find_process_created_process(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                parents = FIND process CREATED procs
                """
        s.execute(stmt)
        data = s.get_variable("parents")
        print(json.dumps(data, indent=4))
        assert len(data)


def test_find_refs_resolution_not_reversed_src_ref(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                nt = get network-traffic
                     from file://{proc_bundle_file}
                     where src_port > 0
                p = FIND process CREATED nt
                """
        s.execute(stmt)
        p = s.get_variable("p")
        # assert len(p) == 948  # grep -c opened_connection_refs tests/doctored-1k.json
        assert len(p) >= 948  # FIXME: duplicate process objects


def test_find_refs_resolution_reversed_src_ref(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where name LIKE '%'
                conns = FIND network-traffic CREATED BY procs
                """
        s.execute(stmt)
        conns = s.get_variable("conns")
        assert (
            len(conns) == 853
        )  # FIXME: should be 948, I think (id collisions for network-traffic)

        # DISP with a ref (parent_ref) and ambiguous column (command_line)
        disp_out = s.execute("DISP procs ATTR name, parent_ref.name, command_line")
        data = disp_out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))


def test_find_with_where_ext_pattern(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                proc1 = FIND process CREATED conns
                        WHERE ipv4-addr:value = '192.168.1.1'

                proc2 = FIND process CREATED conns
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        proc1 = s.symtable["proc1"]
        proc2 = s.symtable["proc2"]

        assert len(conns) == 193
        assert conns.records_count == 203

        assert len(proc1) == 203
        assert proc1.records_count == 203

        assert len(proc2) == 674
        assert proc2.records_count == 674


def test_find_with_where_centered_pattern(proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                procs = FIND process CREATED conns
                        WHERE name = 'vmware.exe'
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 193
        assert conns.records_count == 203

        # FIX ME: It should be 1, not 204
        #         Currently, the prefetch part is good and returns 1,
        #         But the relation resolution does not take WHERE into account
        #         Need solution: store.extract() or store.assign_query()
        assert len(procs) == 204
        assert procs.records_count == 204



"""The STIX-shifter data source package provides access to data sources via
`stix-shifter`_.

The STIX-shifter interface connects to multiple data sources. Users need to
provide one *profile* per data source. The profile name (case insensitive) will
be used in the ``FROM`` clause of the Kestrel ``GET`` command, e.g., ``newvar =
GET entity-type FROM stixshifter://profilename WHERE ...``. Kestrel runtime
will load profiles from 3 places (the later will override the former):

#. STIX-shifter interface config file (only when a Kestrel session starts):

    Create the STIX-shifter interface config file (YAML):

    - Default path: ``~/.config/kestrel/stixshifter.yaml``.

    - A customized path specified in the environment variable ``KESTREL_STIXSHIFTER_CONFIG``.

    Example of STIX-shifter interface config file containing profiles
    (note that the ``options`` section is not required):

    .. code-block:: yaml

        profiles:
            host101:
                connector: elastic_ecs
                connection:
                    host: elastic.securitylog.company.com
                    port: 9200
                    selfSignedCert: false # this means do NOT check cert
                    indices: host101
                    options:  # use any of this section when needed
                        retrieval_batch_size: 10000  # set to 10000 to match default Elasticsearch page size; Kestrel default across connectors: 2000
                        single_batch_timeout: 120  # increase it if hit 60 seconds (Kestrel default) timeout error for each batch of retrieval
                        dialects:  # more info: https://github.com/opencybersecurityalliance/stix-shifter/tree/develop/stix_shifter_modules/elastic_ecs#dialects
                          - beats  # need it if the index is created by Filebeat/Winlogbeat/*beat
                config:
                    auth:
                        id: VuaCfGcBCdbkQm-e5aOx
                        api_key: ui2lp2axTNmsyakw9tvNnw
            host102:
                connector: qradar
                connection:
                    host: qradar.securitylog.company.com
                    port: 443
                config:
                    auth:
                        SEC: 123e4567-e89b-12d3-a456-426614174000
            host103:
                connector: cbcloud
                connection:
                    host: cbcloud.securitylog.company.com
                    port: 443
                config:
                    auth:
                        org-key: D5DQRHQP
                        token: HT8EMI32DSIMAQ7DJM
        options:  # this section is not required
            fast_translate:  # use firepit-native translation (Dataframe as vessel) instead of stix-shifter result translation (JSON as vessel) for the following connectors
                - qradar
                - elastic_ecs

    Full specifications for data source profile sections/fields:

    - Connector-specific fields: in `stix-shifter`_, go to ``stix_shifter_modules/connector_name/configuration`` like `elastic_ecs config`_.

    - General fields shared across connectors: in `stix-shifter`_, go to `stix_shifter_modules/lang_en.json`_.

#. environment variables (only when a Kestrel session starts):

    Three environment variables are required for each profile:

    - ``STIXSHIFTER_PROFILENAME_CONNECTOR``: the STIX-shifter connector name,
      e.g., ``elastic_ecs``.

    - ``STIXSHIFTER_PROFILENAME_CONNECTION``: the STIX-shifter `connection
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#connection>`_
      object in JSON string.

    - ``STIXSHIFTER_PROFILENAME_CONFIG``: the STIX-shifter `configuration
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#configuration>`_
      object in JSON string.

    Example of environment variables for a profile:

    .. code-block:: console

        $ export STIXSHIFTER_HOST101_CONNECTOR=elastic_ecs
        $ export STIXSHIFTER_HOST101_CONNECTION='{"host":"elastic.securitylog.company.com", "port":9200, "indices":"host101"}'
        $ export STIXSHIFTER_HOST101_CONFIG='{"auth":{"id":"VuaCfGcBCdbkQm-e5aOx", "api_key":"ui2lp2axTNmsyakw9tvNnw"}}'

#. any in-session edit through the ``CONFIG`` command.

If you launch Kestrel in debug mode, STIX-shifter debug mode is still not
enabled by default. To record debug level logs of STIX-shifter, create
environment variable ``KESTREL_STIXSHIFTER_DEBUG`` with any value.

.. _STIX-shifter: https://github.com/opencybersecurityalliance/stix-shifter
.. _elastic_ecs config: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/stix_shifter_modules/elastic_ecs/configuration/lang_en.json
.. _stix_shifter_modules/lang_en.json: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/stix_shifter_modules/lang_en.json

"""

import asyncio
import json
import copy
import logging


from stix_shifter.stix_translation import stix_translation
from stix_shifter.stix_transmission import stix_transmission
from stix_shifter_utils.stix_translation.src.utils.transformer_utils import (
    get_module_transformers,
)

from firepit.aio import asyncwrapper
from firepit.aio.ingest import ingest, translate
from kestrel.utils import mkdtemp
from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromStore
from kestrel.exceptions import DataSourceError, DataSourceManagerInternalError
from kestrel_datasource_stixshifter.connector import check_module_availability
from kestrel_datasource_stixshifter.config import (
    get_datasource_from_profiles,
    load_options,
    load_profiles,
    set_stixshifter_logging_level,
)

_logger = logging.getLogger(__name__)


class StixShifterInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX-shifter data source interface only supports ``stixshifter://`` scheme."""
        return ["stixshifter"]

    @staticmethod
    def list_data_sources(config):
        """Get configured data sources from environment variable profiles."""

        # CONFIG command is not supported
        # profiles will be updated according to YAML file and env var
        config["profiles"] = load_profiles()

        data_sources = list(config["profiles"].keys())
        data_sources.sort()
        return data_sources

    @staticmethod
    async def query(uri, pattern, session_id, config, store=None):
        """Query a stixshifter data source."""

        # CONFIG command is not supported
        # profiles will be updated according to YAML file and env var
        config["profiles"] = load_profiles()

        config["options"] = load_options()
        _logger.debug(
            "fast_translate enabled for: %s", config["options"]["fast_translate"]
        )

        scheme, _, profile = uri.rpartition("://")
        profiles = profile.split(",")
        if scheme != "stixshifter":
            raise DataSourceManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        set_stixshifter_logging_level()

        ingestdir = mkdtemp()
        query_id = ingestdir.name
        # bundles = []
        dict_bundles = []
        _logger.debug(f"prepare query with ID: {query_id}")
        for i, profile in enumerate(profiles):
            # STIX-shifter will alter the config objects, thus making them not reusable.
            # So only give STIX-shifter a copy of the configs.
            # Check `modernize` functions in the `stix_shifter_utils` for details.
            (
                connector_name,
                connection_dict,
                configuration_dict,
                retrieval_batch_size,
            ) = map(
                copy.deepcopy, get_datasource_from_profiles(profile, config["profiles"])
            )

            check_module_availability(connector_name)

            data_path_striped = "".join(filter(str.isalnum, profile))

            ingest_stixbundle_filepath = make_ingest_stixbundle_filepath(
                ingestdir, data_path_striped, i
            )

            query_metadata = {
                "id": "identity--" + query_id,
                "name": connector_name,
                "type": "identity",
            }

            translation = stix_translation.StixTranslation()
            transmission = stix_transmission.StixTransmission(
                connector_name, connection_dict, configuration_dict
            )

            translation_options = copy.deepcopy(connection_dict.get("options", {}))
            dsl = await translation.translate_async(
                connector_name, "query", query_metadata, pattern, translation_options
            )

            if "error" in dsl:
                raise DataSourceError(
                    f"STIX-shifter translation from STIX to native query failed with message: {dsl['error']}"
                )

            _logger.debug(f"STIX pattern to query: {pattern}")
            _logger.debug(f"translate results: {dsl}")

            # query results should be put together; when translated to STIX, the relation between them will remain
            transmission_queue = asyncio.Queue()

            # schedule consumers
            consumers = []
            consumer_count = config["options"]["async_translation_workers_count"]
            if connector_name in config["options"]["fast_translate"]:
                for _ in range(consumer_count):
                    consumer = asyncio.create_task(
                        fast_translate_ingest_consume(
                            transmission_queue,
                            connector_name,
                            translation,
                            translation_options,
                            query_metadata,
                            query_id,
                            store,
                        )
                    )
                    consumers.append(consumer)
            else:
                for _ in range(consumer_count):
                    consumer = asyncio.create_task(
                        translation_ingest_consume(
                            transmission_queue,
                            translation,
                            connector_name,
                            query_metadata,
                            translation_options,
                            ingest_stixbundle_filepath,
                            query_id,
                            store,
                        )
                    )
                    consumers.append(consumer)

            batch_index = 0
            for query in dsl["queries"]:
                search_meta_result = await transmission.query_async(query)
                if search_meta_result["success"]:
                    search_id = search_meta_result["search_id"]

                    await transmission_complete(transmission, search_id)

                    # run the producer and wait for completion
                    batch_index = await transmission_produce(
                        connector_name,
                        transmission_queue,
                        transmission,
                        search_id,
                        retrieval_batch_size,
                        batch_index,
                    )

                else:
                    stix_shifter_error_msg = (
                        search_meta_result["error"]
                        if "error" in search_meta_result
                        else "details not avaliable"
                    )
                    raise DataSourceError(
                        f"STIX-shifter transmission.query() failed with message: {stix_shifter_error_msg}"
                    )

            # _logger.debug("transmission succeeded, start translate back to STIX")

            # wait until the consumer has processed all items
            await transmission_queue.join()

            # the consumers are still waiting for an item, cancel them
            for consumer in consumers:
                consumer.cancel()

            # wait until all worker tasks are cancelled
            responses = await asyncio.gather(*consumers, return_exceptions=True)
            for response in responses:
                if not isinstance(response, asyncio.CancelledError):
                    raise response

        return ReturnFromStore(query_id)
        # return ReturnFromFile(query_id, bundles)


def make_ingest_stixbundle_filepath(ingestdir, data_path_striped, profile_index):
    def ingest_stixbundle_filepath(batch_index):
        ingestbatchfile = (
            ingestdir / f"{profile_index}_{batch_index}_{data_path_striped}.json"
        )
        return ingestbatchfile

    return ingest_stixbundle_filepath


async def transmission_complete(transmission, search_id):
    status = None
    while True:
        status = await transmission.status_async(search_id)
        if not status["success"]:
            stix_shifter_error_msg = (
                status["error"] if "error" in status else "details not avaliable"
            )
            raise DataSourceError(
                f"STIX-shifter transmission.status() failed with message: {stix_shifter_error_msg}"
            )
        elif status["progress"] < 100 and status["status"] == "RUNNING":
            await asyncio.sleep(1)
        else:
            break


async def transmission_produce(
    connector_name,
    transmission_queue,
    transmission,
    search_id,
    retrieval_batch_size,
    batch_index,
):
    result_retrieval_offset = 0
    has_remaining_results = True
    metadata = None
    is_retry_cycle = False
    while has_remaining_results:
        result_batch = await transmission.results_async(
            search_id, result_retrieval_offset, retrieval_batch_size, metadata
        )
        if result_batch["success"]:
            new_entries = result_batch["data"]
            if new_entries:
                result_batch["batch_index"] = batch_index
                await transmission_queue.put(result_batch)
                result_retrieval_offset += len(new_entries)
                batch_index += 1
            else:
                has_remaining_results = False
            if "metadata" in result_batch:
                metadata = result_batch["metadata"]
            is_retry_cycle = False
        else:
            stix_shifter_error_msg = (
                result_batch["error"]
                if "error" in result_batch
                else "details not avaliable"
            )
            if (
                stix_shifter_error_msg.startswith(
                    f"{connector_name} connector error => server timeout_error"
                )
                and not is_retry_cycle
            ):
                # mitigate https://github.com/opencybersecurityalliance/stix-shifter/issues/1493
                # only give it one retry to mitigate high CPU occupation
                # otherwise, it could be a real server connection issue
                # /stix_shifter_utils/stix_transmission/utils/RestApiClientAsync.py
                _logger.info(
                    f"busy CPU; hit stix-shifter transmission aiohttp connection timeout; retry"
                )
                is_retry_cycle = True
            else:
                raise DataSourceError(
                    f"STIX-shifter transmission.results() failed with message: {stix_shifter_error_msg}"
                )
    return batch_index


async def translation_ingest_consume(
    transmission_queue,
    translation,
    connector_name,
    query_metadata,
    translation_options,
    ingest_stixbundle_filepath,
    query_id,
    store,
):
    while True:
        # wait for an item from the producer
        result_batch = await transmission_queue.get()

        stixbundle = await translation.translate_async(
            connector_name,
            "results",
            query_metadata,
            result_batch["data"],
            translation_options,
        )

        if "error" in stixbundle:
            transmission_queue.task_done()
            raise DataSourceError(
                f"STIX-shifter translation results to STIX failed with message: {stixbundle['error']}"
            )

        await asyncwrapper.SyncWrapper(store=store).cache(query_id, stixbundle)

        if _logger.isEnabledFor(logging.DEBUG):
            ingestbatchfile = ingest_stixbundle_filepath(
                str(result_batch["batch_index"])
            )
            _logger.debug(f"dumping STIX bundles into file: {ingestbatchfile}")
            with ingestbatchfile.open("w") as ingest_fp:
                json.dump(stixbundle, ingest_fp, indent=4)

        # Notify the queue that the item has been processed
        transmission_queue.task_done()


async def fast_translate_ingest_consume(
    transmission_queue,
    connector_name,
    translation,
    translation_options,
    identity,
    query_id,
    store,
):
    while True:
        # wait for an item from the producer
        result_batch = await transmission_queue.get()
        await fast_translate(
            connector_name,
            result_batch["data"],
            translation,
            translation_options,
            identity,
            query_id,
            store,
        )
        # Notify the queue that the item has been processed
        transmission_queue.task_done()


async def fast_translate(
    connector_name,
    connector_results,
    translation,
    translation_options,
    identity,
    query_id,
    store,
):
    # Use the alternate, faster DataFrame-based translation (in firepit)
    _logger.debug("Using fast translation for connector %s", connector_name)
    transformers = get_module_transformers(connector_name)
    mapping = await translation.translate_async(
        connector_name,
        stix_translation.MAPPING,
        None,
        None,
        translation_options,
    )

    if "error" in mapping:
        raise DataSourceError(
            f"STIX-shifter mapping failed with message: {mapping['error']}"
        )

    df = translate(mapping["to_stix_map"], transformers, connector_results, identity)

    identity_obj = {
        "identity_class": "system",
        "created": None,
        "modified": None,
    }  # These are required by STIX but not needed here
    identity_obj.update(identity)

    await ingest(
        asyncwrapper.SyncWrapper(store=store),
        identity_obj,
        df,
        query_id,
    )



import asyncio

from stix_shifter.stix_translation import stix_translation

from kestrel.session import Session
from kestrel_datasource_stixshifter.connector import check_module_availability
from kestrel_datasource_stixshifter.interface import fast_translate


SAMPLE_RESULT = {
    "success": True,
    "data": [
        {
            "process": {
                "name": "svchost.exe",
                "pid": 1744,
                "entity_id": "{8dfc401c-f5a9-6068-2300-000000001300}",
                "executable": "C:\\Windows\\System32\\svchost.exe",
            },
            "@timestamp": "2021-04-10T05:35:45.333Z",
            "destination": {
                "port": 49343,
                "ip": "127.0.0.1",
                "domain": "thl-win.example.com",
            },
            "host": {
                "hostname": "thl-win",
                "os": {
                    "build": "17763.1817",
                    "kernel": "10.0.17763.1817 (WinBuild.160101.0800)",
                    "name": "Windows Server 2019 Standard",
                    "family": "windows",
                    "version": "10.0",
                    "platform": "windows",
                },
                "ip": ["2001:db8:a::123", "198.51.100.101", "10.0.0.141"],
                "name": "thl-win.example.com",
                "id": "d3f820f9-f601-4dd8-bd48-5c13549a6027",
                "mac": ["00:25:96:12:34:56"],
                "architecture": "x86_64",
            },
            "source": {
                "port": 49343,
                "ip": "127.0.0.1",
                "domain": "thl-win.example.com",
            },
            "event": {
                "code": 3,
                "provider": "Microsoft-Windows-Sysmon",
                "kind": "event",
                "created": "2021-04-10T05:35:43.069Z",
                "module": "sysmon",
                "action": "Network connection detected (rule: NetworkConnect)",
                "category": ["network"],
                "type": ["connection", "start", "protocol"],
            },
            "user": {"domain": "NT AUTHORITY", "name": "SYSTEM"},
            "network": {
                "community_id": "1:ZXhhbXBsZQ==",
                "protocol": "-",
                "transport": "udp",
                "type": "ipv4",
                "direction": "egress",
            },
            "tags": ["beats_input_codec_plain_applied"],
        }
    ],
}


def test_fast_translate():
    connector_name = "elastic_ecs"
    check_module_availability(connector_name)
    connector_results = SAMPLE_RESULT["data"]
    translation = stix_translation.StixTranslation()
    translation_options = {}
    query_id = "8df266aa-2901-4a94-ace9-a4403e310fa1"
    identity = {"id": "identity--" + query_id, "name": connector_name}
    with Session() as s:
        asyncio.run(
            fast_translate(
                connector_name,
                connector_results,
                translation,
                translation_options,
                identity,
                query_id,
                s.store,
            )
        )



import os
import pandas as pd
import pytest

from kestrel.exceptions import VariableNotExist
from kestrel.session import Session


@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "doctored-1k.json")


def test_disp():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        d = s.execute(stmt)
        out = s.execute("DISP newvar")
        data = out[0].to_dict()


def test_disp_no_vars():
    with Session() as s:
        with pytest.raises(VariableNotExist):
            s.execute("DISP newvar")


def test_disp_grouped_procs():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             , {"type": "process", "name": "explorer.exe", "pid": "98"}
             ]
"""
        s.execute(stmt)
        s.execute("grpvar = group newvar by name")
        out = s.execute("DISP grpvar")
        data = out[0].to_dict()["data"]
        assert len(data) == 2


def test_disp_grouped_conns():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "network-traffic", "src_ref.value": "1.2.3.4", "dst_ref.value": "4.3.2.1"}
             , {"type": "network-traffic", "src_ref.value": "2.2.3.4", "dst_ref.value": "4.3.2.1"}
             , {"type": "network-traffic", "src_ref.value": "1.2.3.4", "dst_ref.value": "5.3.2.1"}
             ]
"""
        s.execute(stmt)
        s.execute("grpvar = group newvar by dst_ref.value")
        out = s.execute("DISP grpvar")
        data = out[0].to_dict()["data"]
        assert len(data) == 2


def test_disp_mixed_v4_v6(proc_bundle_file):
    with Session() as s:
        stmt = f"""
conns = GET network-traffic
        FROM file://{proc_bundle_file}
        WHERE [network-traffic:dst_port > 0]
"""
        s.execute(stmt)

        out = s.execute("DISP conns ATTR src_ref.value, src_port")
        data = out[0].to_dict()["data"]
        df = pd.DataFrame.from_records(data)
        assert df.columns.tolist() == ["src_ref.value", "src_port"]

        out = s.execute("DISP TIMESTAMPED(conns) ATTR src_ref.value, src_port")
        data = out[0].to_dict()["data"]
        df = pd.DataFrame.from_records(data)
        assert df.columns.tolist() == ["first_observed", "src_ref.value", "src_port"]



import re

from lark import UnexpectedToken
import pytest

from kestrel.syntax.parser import parse_kestrel, parse_ecgpattern
from kestrel.syntax.ecgpattern import Reference
from kestrel.exceptions import InvalidECGPattern
from firepit.timestamp import timefmt


@pytest.mark.parametrize("pattern",
    ["[url:value LIKE '%']"
    ,"url:value LIKE '%'"
    ,"value LIKE '%'"
    ,"[value LIKE '%']"
    ,"[url:value \n LIKE '%']"
    ,"value LIKE \n '%'"
    ],
)
def test_simple_get(pattern):
    results = parse_kestrel(f"y = get url from udi://all where {pattern}")
    result = results[0]

    assert result["command"] == "get"
    assert result["type"] == "url"
    assert result["datasource"] == "udi://all"

    where = results[0]["where"]
    where.add_center_entity("url")
    assert where.to_stix(None, None) == "[url:value LIKE '%']"


@pytest.mark.parametrize("pattern",
    ["pid IN (1, 2, 3)"
    ,"[pid IN (1, 2, 3)]"
    ,"[process:pid IN (1, 2, 3)]"
    ],
)
def test_assign_in(pattern):
    results = parse_kestrel(f"y = x WHERE {pattern}")
    where = results[0]["where"]
    where.add_center_entity("process")
    assert where.to_stix(None, None) == "[process:pid IN (1,2,3)]"


@pytest.mark.parametrize(
    "pattern, errprint",
    [
        ("pid = (1, 2, 3)", 'a list should be paired with the operator "IN"'),
        ("pid IN 'asdf'", 'inappropriately pair operator "IN" with literal'),
    ],
)
def test_ecgp_in_exception(pattern, errprint):
    with pytest.raises(InvalidECGPattern) as einfo:
        results = parse_kestrel(f"y = x WHERE {pattern}")
    assert einfo.value.error == errprint


def test_quoted_datasource():
    results = parse_kestrel("y = get url from \"udi://My QRadar\" where [url:value LIKE '%']")
    result = results[0]
    assert result["command"] == "get"
    assert result["type"] == "url"
    assert result["datasource"] == "udi://My QRadar"

    where = results[0]["where"]
    where.add_center_entity("url")
    assert where.to_stix(None, None) == "[url:value LIKE '%']"


@pytest.mark.parametrize(
    "ecgp, center_entity, stix",
    [
        (
            "name = 'powershell.exe'",
            "process",
            "[process:name = 'powershell.exe']",
        ),
        (
            "name = 'powershell.exe' AND pid = 1234",
            "process",
            "[(process:name = 'powershell.exe' AND process:pid = 1234)]",
        ),
        (
            r"name = 'power\'xyz\'.exe'",
            r"process",
            r"[process:name = 'power\'xyz\'.exe']",
        ),
        (
            r"""name = 'po\'wer"xyz\".exe'""",
            r"""process""",
            r"""[process:name = 'po\'wer"xyz".exe']""",
        ),
        (
            r"""name = "po'wer\"xyz\".exe" """,
            r"""process""",
            r"""[process:name = 'po\'wer"xyz".exe']""",
        ),
        (
            r"command_line = 'C:\\abc\\xyz.exe /c asdf'",
            r"process",
            r"[process:command_line = 'C:\\abc\\xyz.exe /c asdf']",
        ),
        (
            r"command_line = r'C:\abc\xyz.exe /c asdf'",
            r"process",
            r"[process:command_line = 'C:\\abc\\xyz.exe /c asdf']",
        ),
        (
            r'command_line = r"C:\wbc\nyz.exe /c asdf"',
            r"process",
            r"[process:command_line = 'C:\\wbc\\nyz.exe /c asdf']",
        ),
        (
            "name LIKE 'power%.exe'",
            "process",
            "[process:name LIKE 'power%.exe']",
        ),
        (
            r"name MATCHES 'power.+\\d{1,3}[a-zA-Z0-9]+\\.exe'",
            r"process",
            r"[process:name MATCHES 'power.+\\d{1,3}[a-zA-Z0-9]+\\.exe']",
        ),
        (
            r"name MATCHES 'power\\(hi\\)\\w+(real)\\.exe'",
            r"process",
            r"[process:name MATCHES 'power\\(hi\\)\\w+(real)\\.exe']",
        ),
        (
            r"name MATCHES 'C:\\\\Windows\\\\system32\\\\svchost\\.exe'",
            r"process",
            r"[process:name MATCHES 'C:\\\\Windows\\\\system32\\\\svchost\\.exe']",
        ),
        (
            r"name MATCHES r'C:\\Windows\\system32\\svchost\.exe'",
            r"process",
            r"[process:name MATCHES 'C:\\\\Windows\\\\system32\\\\svchost\\.exe']",
        ),
    ],
)
def test_ecgp(ecgp, center_entity, stix):
    # test ECGP in GET
    stmt = f"x = GET {center_entity} FROM xxx WHERE {ecgp}"
    cmd = parse_kestrel(stmt)[0]
    cmd["where"].add_center_entity(cmd["type"])
    assert cmd["where"].to_stix(None, None) == stix

    # test ECGP for standalone parsing
    pattern = parse_ecgpattern(ecgp)
    pattern.add_center_entity(center_entity)
    assert pattern.to_stix(None, None) == stix


@pytest.mark.parametrize(
    "outvar, sco_type, ds, pat",
    [
        ("_my_var", "ipv4-addr", "something", "[ipv4-addr:value = '192.168.121.121']"),
        (
            "X1",
            "x-custom-object",
            "myscheme://foo.bar/whatever",
            "[x-other-custom-thing:x_custom_prop IN ('a','b','c')]",
        ),
        (
            "urls",
            "url",
            "file:///shared-vol/udsstx",
            "[url:value LIKE 'https://%'] START t'2021-03-29T19:25:12.345Z' STOP t'2021-03-29T19:30:12.345Z'",
        ),
        (
            "ext_dns_conns",
            "network-traffic",
            '"udi://10k Traffic"',
            "[(network-traffic:dst_port = 53 AND network-traffic:dst_ref.value NOT ISSUBSET '192.168.1.0/24')]",
        ),
    ],
)
def test_get(outvar, sco_type, ds, pat):
    results = parse_kestrel(f"{outvar} = GET {sco_type} FROM {ds} WHERE {pat}")
    result = results[0]
    assert result["output"] == outvar
    assert result["command"] == "get"
    assert result["type"] == sco_type
    assert result["datasource"] == ds.strip('"')  # We strip the double quotes

    where = results[0]["where"]
    where.add_center_entity("sco_type")
    assert where.to_stix(result["timerange"], None) == pat


def test_get_timerange():
    results = parse_kestrel("""
        x = GET process
            FROM xxx
            WHERE name = 'asdf'
            START 2022-10-18T01:02:03Z
            STOP  2022-10-19T04:05:06Z
        """)
    result = results[0]
    assert timefmt(result["timerange"][0]) == "2022-10-18T01:02:03.000Z"
    assert timefmt(result["timerange"][1]) == "2022-10-19T04:05:06.000Z"

    stix = "[process:name = 'asdf'] START t'2022-10-18T01:02:03.000Z' STOP t'2022-10-19T04:05:06.000Z'"
    result["where"].add_center_entity(result["type"])
    assert result["where"].to_stix(result["timerange"], None) == stix


def test_apply_params():
    results = parse_kestrel("apply xyz://my_analytic on foo with x=1, y=(a, b,c)")
    result = results[0]
    assert result["command"] == "apply"
    assert result["analytics_uri"] == "xyz://my_analytic"
    assert result["inputs"] == ["foo"]
    assert result["arguments"] == {"x": 1, "y": ["a", "b", "c"]}


def test_apply_params_with_dots():
    results = parse_kestrel("apply xyz://my_analytic on foo with x=0.1, y=a.value")
    result = results[0]
    assert result["command"] == "apply"
    assert result["analytics_uri"] == "xyz://my_analytic"
    assert result["inputs"] == ["foo"]
    assert result["arguments"] == {"x": 0.1, "y": Reference("a", "value")}


def test_apply_params_with_at():
    results = parse_kestrel("apply xyz://my_analytic on foo with x=\"https://www.xyz.com/123@me.com/action\"")
    result = results[0]
    assert result["command"] == "apply"
    assert result["analytics_uri"] == "xyz://my_analytic"
    assert result["inputs"] == ["foo"]
    assert result["arguments"] == {"x": "https://www.xyz.com/123@me.com/action"}


def test_apply_params_with_decimal_and_dots():
    results = parse_kestrel("apply xyz://my_analytic on foo with x=0.1, y=[a.value ,b,c]")
    result = results[0]
    assert result["command"] == "apply"
    assert result["analytics_uri"] == "xyz://my_analytic"
    assert result["inputs"] == ["foo"]
    assert result["arguments"] == {"x": 0.1, "y": [Reference("a", "value"), "b", "c"]}


def test_apply_params_no_equals():
    with pytest.raises(UnexpectedToken):
        parse_kestrel("apply xyz://my_analytic on foo with x=1, y")


def test_grouping_0():
    results = parse_kestrel("y = group x by foo")
    result = results[0]
    print(result)
    assert result["command"] == "group"
    assert result["input"] == "x"
    assert result["attributes"] == ["foo"]
    assert "aggregations" not in result


def test_grouping_1():
    results = parse_kestrel("y = group x by foo with sum(baz)")
    result = results[0]
    print(result)
    assert result["command"] == "group"
    assert result["input"] == "x"
    assert result["attributes"] == ["foo"]
    assert result["aggregations"] == [
        {"attr": "baz", "func": "sum", "alias": "sum_baz"},
    ]


def test_grouping_2():
    results = parse_kestrel("y = group x BY foo, bar WITH MAX(baz) AS biggest, MIN(blah)")
    result = results[0]
    print(result)
    assert result["command"] == "group"
    assert result["input"] == "x"
    assert result["attributes"] == ["foo", "bar"]
    assert result["aggregations"] == [
        {"attr": "baz", "func": "max", "alias": "biggest"},
        {"attr": "blah", "func": "min", "alias": "min_blah"},
    ]


def test_grouping_3():
    results = parse_kestrel(
        "y = group x by foo with avg(bar), count(baz), max(blah) as whatever"
    )
    result = results[0]
    print(result)
    assert result["command"] == "group"
    assert result["input"] == "x"
    assert result["attributes"] == ["foo"]
    assert result["aggregations"] == [
        {"attr": "bar", "func": "avg", "alias": "avg_bar"},
        {"attr": "baz", "func": "count", "alias": "count_baz"},
        {"attr": "blah", "func": "max", "alias": "whatever"},
    ]



import pytest
import os

from kestrel.session import Session

from kestrel_datasource_stixshifter.connector import (
    verify_package_origin,
    check_module_availability,
)

from kestrel_datasource_stixshifter.config import get_datasource_from_profiles


def test_verify_package_origin():
    connectors = ["stix_bundle", "qradar", "elastic_ecs", "splunk"]
    for connector_name in connectors:
        verify_package_origin(connector_name, "test_version")


def test_check_module_availability():
    connectors = ["stix_bundle"]
    for connector_name in connectors:
        check_module_availability(connector_name)


def test_yaml_profiles_refresh(tmp_path):

    profileA = f"""profiles:
profiles:
    host101:
        connector: elastic_ecs
        connection:
            host: elastic.securitylog.company.com
            port: 9200
            indices: host101
        config:
            auth:
                id: profileA
                api_key: qwer
"""

    profileB = f"""profiles:
profiles:
    host101:
        connector: elastic_ecs
        connection:
            host: elastic.securitylog.company.com
            port: 9200
            indices: host101
            options:
                retrieval_batch_size: 10000
                single_batch_timeout: 120
                dialects:
                    - beats
        config:
            auth:
                id: profileB
                api_key: asdf
"""

    profile_file = tmp_path / "stixshifter.yaml"

    os.environ["KESTREL_STIXSHIFTER_CONFIG"] = str(profile_file.expanduser().resolve())

    with Session() as s:

        with open(profile_file, "w") as pf:
            pf.write(profileA)

        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        s.execute(stmt)

        s.data_source_manager.list_data_sources_from_scheme("stixshifter")

        ss_config = s.config["datasources"]["kestrel_datasource_stixshifter"]
        ss_profiles = ss_config["profiles"]
        connector_name, connection, configuration, retrieval_batch_size = get_datasource_from_profiles("host101", ss_profiles)
        assert connector_name == "elastic_ecs"
        assert configuration["auth"]["id"] == "profileA"
        assert configuration["auth"]["api_key"] == "qwer"
        assert connection["options"]["timeout"] == 60
        assert connection["options"]["result_limit"] == 2000 * 2
        assert retrieval_batch_size == 2000

        with open(profile_file, "w") as pf:
            pf.write(profileB)

        s.data_source_manager.list_data_sources_from_scheme("stixshifter")

        # need to refresh the pointers since the dict is updated
        ss_profiles = ss_config["profiles"]
        connector_name, connection, configuration, retrieval_batch_size = get_datasource_from_profiles("host101", ss_profiles)
        assert connector_name == "elastic_ecs"
        assert configuration["auth"]["id"] == "profileB"
        assert configuration["auth"]["api_key"] == "asdf"
        assert connection["options"]["timeout"] == 120
        assert connection["options"]["result_limit"] == 10000 * 2
        assert retrieval_batch_size == 10000



import pytest
import pathlib
import os

from kestrel.session import Session
from kestrel.codegen.display import DisplayHtml


NEW_PROCS = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""


REF_PROCS = """
ref = NEW [
          {"type": "process", "name": "", "pid": 4},
          {"type": "process", "name": "explorer.exe", "pid": 1380}
          ]
"""


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


@pytest.fixture
def fake_bundle_4():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle_4.json")


@pytest.fixture(autouse=True)
def env_setup(tmp_path):

    analytics_module_path = str(
        pathlib.Path(__file__).resolve().parent / "python_analytics_mockup.py"
    )

    profiles = f"""profiles:
    enrich_one_variable:
        module: {analytics_module_path}
        func: enrich_one_variable
    html_visualization:
        module: {analytics_module_path}
        func: html_visualization
    enrich_multiple_variables:
        module: {analytics_module_path}
        func: enrich_multiple_variables
    enrich_variable_with_arguments:
        module: {analytics_module_path}
        func: enrich_variable_with_arguments
    """

    profile_file = tmp_path / "pythonanalytics.yaml"
    with open(profile_file, "w") as pf:
        pf.write(profiles)

    os.environ["KESTREL_PYTHON_ANALYTICS_CONFIG"] = str(
        profile_file.expanduser().resolve()
    )


def test_enrich_one_variable():
    with Session() as s:
        s.execute(NEW_PROCS)
        s.execute("APPLY python://enrich_one_variable ON newvar")
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert set([v[0]["x_new_attr"], v[1]["x_new_attr"]]) == set(
            ["newval0", "newval1"]
        )


def test_html_visualization():
    with Session() as s:
        s.execute(NEW_PROCS)
        displays = s.execute("APPLY python://html_visualization ON newvar")
        viz = displays[0]
        assert type(viz) is DisplayHtml
        assert viz.html == "<p>Hello World! -- a Kestrel analytics</p>"


def test_enrich_multiple_variables():
    with Session() as s:
        stmt = """
               v1 = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
                        , {"type": "process", "name": "explorer.exe", "pid": "99"}
                        ]
               v2 = NEW process ["cmd.exe", "explorer.exe", "google-chrome.exe"]
               v3 = NEW ipv4-addr ["1.1.1.1", "2.2.2.2"]
               APPLY python://enrich_multiple_variables ON v1, v2, v3
               """
        s.execute(stmt)

        v1 = s.get_variable("v1")
        assert set([v1[0]["x_new_attr"], v1[1]["x_new_attr"]]) == set(
            ["newval_a0", "newval_a1"]
        )

        v2 = s.get_variable("v2")
        assert set(
            [v2[0]["x_new_attr"], v2[1]["x_new_attr"], v2[2]["x_new_attr"]]
        ) == set(["newval_b0", "newval_b1", "newval_b2"])

        v3 = s.get_variable("v3")
        assert set([v3[0]["x_new_attr"], v3[1]["x_new_attr"]]) == set(
            ["newval_c0", "newval_c1"]
        )


def test_enrich_variable_with_arguments():
    with Session() as s:
        s.execute(NEW_PROCS)
        stmt = """
               APPLY python://enrich_variable_with_arguments
                     ON newvar
                     WITH argx=(abc,888), argy=123, argz="as\\"is"
               """
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["x_new_argx"] == "abc,888"
        assert v[0]["x_new_argy"] == '123'
        assert v[0]["x_new_argz"] == r'as"is'


def test_enrich_variable_with_reference_in_arguments():
    with Session() as s:
        s.execute(NEW_PROCS)
        s.execute(REF_PROCS)
        stmt = """
               APPLY python://enrich_variable_with_arguments
                     ON newvar
                     WITH argx=ref.pid, argy=123, argz=[ref.pid, 555]
               """
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 2
        assert v[0]["type"] == "process"
        assert v[0]["x_new_argx"] in ("4,1380", "1380,4")
        assert v[0]["x_new_argy"] == '123'
        assert v[0]["x_new_argz"] in ("4,1380,555", "1380,4,555")



def test_enrich_after_get_url(fake_bundle_file):
    with Session() as s:
        stmt = f"""
                newvar = get url
                         from file://{fake_bundle_file}
                         where value LIKE '%'
                APPLY python://enrich_one_variable ON newvar
                """
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 31
        assert v[0]["type"] == "url"
        assert "x_new_attr" in v[0]


def test_enrich_after_get_process(fake_bundle_4):
    with Session() as s:
        stmt = f"""
               newvar = get process
                        from file://{fake_bundle_4}
                        where binary_ref.name LIKE "%"
               APPLY python://enrich_one_variable ON newvar
               """
        s.execute(stmt)
        v = s.get_variable("newvar")
        assert len(v) == 4
        assert v[0]["type"] == "process"
        assert "x_new_attr" in v[0]



from datetime import datetime, timedelta
from pkgutil import get_data
import importlib
from lark import Lark, Token, Transformer
from lark.visitors import merge_transformers

from firepit.timestamp import to_datetime
from firepit.query import BinnedColumn
from kestrel.utils import unescape_quoted_string, resolve_path
from kestrel.syntax.utils import resolve_uri
from kestrel.syntax.ecgpattern import (
    ECGPComparison,
    ECGPJunction,
    ExtCenteredGraphPattern,
    Reference,
)

DEFAULT_VARIABLE = "_"
DEFAULT_SORT_ORDER = "DESC"


def parse_kestrel(
    stmts, default_variable=DEFAULT_VARIABLE, default_sort_order=DEFAULT_SORT_ORDER
):
    # the public parsing interface for Kestrel
    # return abstract syntax tree
    # check kestrel.lark for details
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    return Lark(
        grammar,
        parser="lalr",
        transformer=_KestrelT(default_variable, default_sort_order),
    ).parse(stmts)


def parse_ecgpattern(pattern_str) -> ExtCenteredGraphPattern:
    grammar = get_data(__name__, "ecgpattern.lark").decode("utf-8")
    paths = importlib.util.find_spec("kestrel.syntax").submodule_search_locations
    return Lark(
        grammar,
        parser="lalr",
        import_paths=paths,
        transformer=merge_transformers(
            _ECGPatternT(), kestrel=_KestrelT(token_prefix="kestrel__")
        ),
    ).parse(pattern_str)


class _ECGPatternT(Transformer):
    def start(self, args):
        return ExtCenteredGraphPattern(args[0])


class _KestrelT(Transformer):
    def __init__(
        self,
        default_variable=DEFAULT_VARIABLE,
        default_sort_order=DEFAULT_SORT_ORDER,
        token_prefix="",
    ):
        # token_prefix is the modification by Lark when using `merge_transformers()`
        self.default_variable = default_variable
        self.default_sort_order = default_sort_order
        self.token_prefix = token_prefix
        super().__init__()

    def start(self, args):
        return args

    def statement(self, args):
        # Kestrel syntax: a statement can only has one command
        stmt = args.pop()
        return stmt

    def assignment(self, args):
        if len(args) > 2:
            variables = self._extract_vars(args)
            stmt = {"command": "merge", "output": variables[0], "inputs": variables[1:]}
        elif len(args) == 2:
            stmt = args[1]
            # the tree is already processed, and the only thing left is the result variable
            # get it using `self._extract_var()`
            stmt["output"] = self._extract_var(args)
            if "command" not in stmt:
                stmt["command"] = "assign"
        else:
            stmt = args[0]
            stmt["output"] = self.default_variable
        return stmt

    def info(self, args):
        return {"command": "info", "input": self._extract_var(args)}

    def disp(self, args):
        packet = {"command": "disp"}
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        if "attrs" not in packet:
            packet["attrs"] = "*"
        return packet

    def get(self, args):
        packet = {
            "command": "get",
            "type": self._extract_entity_type(args),
        }

        for item in args:
            if isinstance(item, dict):
                packet.update(item)

        if "timerange" not in packet:
            packet["timerange"] = None

        return packet

    def find(self, args):
        packet = {
            "command": "find",
            "type": self._extract_entity_type(args),
            "relation": self._assert_and_extract_single("RELATION", args).lower(),
            "reversed": self._extract_if_reversed(args),
            "input": self._extract_var(args),
        }

        for item in args:
            if isinstance(item, dict):
                packet.update(item)

        if "timerange" not in packet:
            packet["timerange"] = None

        return packet

    def join(self, args):
        packet = {
            "command": "join",
            "input": _first(args),
            "input_2": _second(args),
        }
        if len(args) == 5:
            packet["attribute_1"] = _fourth(args)
            packet["attribute_2"] = _fifth(args)

        return packet

    def group(self, args):
        packet = {
            "command": "group",
            "attributes": args[2],
            "input": self._extract_var(args),
        }
        aggregations = args[3] if len(args) > 3 else None
        if aggregations:
            packet["aggregations"] = aggregations
        return packet

    def sort(self, args):
        return {
            "command": "sort",
            "attribute": self._extract_attribute(args),
            "input": self._extract_var(args),
            "ascending": self._extract_direction(args),
        }

    def apply(self, args):
        packet = {"command": "apply", "arguments": {}}
        for arg in args:
            if isinstance(arg, dict):
                if "variables" in arg:
                    packet["inputs"] = arg["variables"]
                else:
                    packet.update(arg)
        return packet

    def load(self, args):
        packet = {
            "command": "load",
            "type": self._extract_entity_type(args),
        }
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        return packet

    def save(self, args):
        packet = {
            "command": "save",
            "input": self._extract_var(args),
        }
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        return packet

    def new(self, args):
        return {
            "command": "new",
            "type": self._extract_entity_type(args),
            "data": self._assert_and_extract_single("VAR_DATA", args),
        }

    def expression(self, args):
        packet = args[0]
        for arg in args:
            packet.update(arg)
        return packet

    def vtrans(self, args):
        return {
            "input": self._extract_var(args),
            "transform": self._assert_and_extract_single("TRANSFORM", args),
        }

    def where_clause(self, args):
        pattern = ExtCenteredGraphPattern(args[0])
        return {
            "where": pattern,
        }

    def expression_or(self, args):
        return ECGPJunction("OR", args[0], args[1])

    def expression_and(self, args):
        return ECGPJunction("AND", args[0], args[1])

    def comparison_std(self, args):
        etype, attr = _extract_entity_and_attribute(args[0].value)
        # remove more than one spaces; capitalize op
        op = " ".join(_second(args).split()).upper()
        value = args[2]
        return ECGPComparison(attr, op, value, etype)

    def comparison_null(self, args):
        etype, attr = _extract_entity_and_attribute(args[0].value)
        op = _second(args)
        if "NOT" in op:
            op = "!="
        else:
            op = "="
        value = "NULL"
        return ECGPComparison(attr, op, value, etype)

    def value(self, args):
        return args[0]

    def literal_list(self, args):
        if len(args) == 1:
            return args[0]
        else:
            return args

    def literal(self, args):
        return args[0]

    def reference_or_simple_string(self, args):
        if len(args) > 1:
            variable = _first(args)
            attribute = _second(args)
            v = Reference(variable, attribute)
        else:
            v = _first(args)
        return v

    def string(self, args):
        raw = _first(args)
        if args[0].type == self.token_prefix + "SIMPLE_STRING":
            value = raw
        elif args[0].type == self.token_prefix + "ADVANCED_STRING":
            value = unescape_quoted_string(raw)
        return value

    def number(self, args):
        v = _first(args)
        try:
            return int(v)
        except:
            return float(v)

    def attr_clause(self, args):
        paths = self._assert_and_extract_single("ATTRIBUTES", args)
        return {
            "attrs": paths if paths else "*",
        }

    def sort_clause(self, args):
        return {
            "attribute": self._extract_attribute(args),
            "ascending": self._extract_direction(args),
        }

    def limit_clause(self, args):
        return {
            "limit": int(_first(args)),
        }

    def offset_clause(self, args):
        return {
            "offset": int(_first(args)),
        }

    def timespan_relative(self, args):
        num = int(args[0])
        unit = args[1]
        if unit.type == self.token_prefix + "DAY":
            delta = timedelta(days=num)
        elif unit.type == self.token_prefix + "HOUR":
            delta = timedelta(hours=num)
        elif unit.type == self.token_prefix + "MINUTE":
            delta = timedelta(minutes=num)
        elif unit.type == self.token_prefix + "SECOND":
            delta = timedelta(seconds=num)
        stop = datetime.utcnow()
        start = stop - delta
        return {"timerange": (start, stop)}

    def timespan_absolute(self, args):
        start = to_datetime(args[0])
        stop = to_datetime(args[1])
        return {"timerange": (start, stop)}

    def timestamp(self, args):
        return self._assert_and_extract_single("ISOTIMESTAMP", args)

    def entity_type(self, args):
        return _first(args)

    def variables(self, args):
        return {"variables": self._extract_vars(args)}

    def stdpath(self, args):
        v = _first(args)
        if args[0].type == self.token_prefix + "PATH_ESCAPED":
            v = unescape_quoted_string(v)
        v = resolve_path(v)
        return {"path": v}

    def datasource(self, args):
        v = _first(args)
        if args[0].type == self.token_prefix + "DATASRC_ESCAPED":
            v = unescape_quoted_string(v)
        v = ",".join(map(resolve_uri, v.split(",")))
        return {"datasource": v}

    def analytics_uri(self, args):
        v = _first(args)
        if args[0].type == self.token_prefix + "ANALYTICS_ESCAPED":
            v = unescape_quoted_string(v)
        return {"analytics_uri": v}

    # automatically put one or more grp_expr into a list
    def grp_spec(self, args):
        return args

    def grp_expr(self, args):
        item = args[0]
        if isinstance(item, Token):
            # an ATTRIBUTE
            return str(item)
        else:
            # bin_func
            return item

    def bin_func(self, args):
        attr = _first(args)
        num = int(_second(args))
        if len(args) >= 3:
            unit = _third(args)
        else:
            unit = None
        alias = f"{attr}_bin"
        return BinnedColumn(attr, num, unit, alias=alias)

    def agg_list(self, args):
        return [arg for arg in args]

    def agg(self, args):
        func = _first(args).lower()
        attr = _second(args)
        alias = _third(args) if len(args) > 2 else f"{func}_{attr}"
        return {"func": func, "attr": attr, "alias": alias}

    def args(self, args):
        d = {}
        for di in args:
            d.update(di)
        return {"arguments": d}

    def arg_kv_pair(self, args):
        return {_first(args): args[1]}

    def _extract_vars(self, args):
        var_names = []
        for arg in args:
            if hasattr(arg, "type") and arg.type == self.token_prefix + "VARIABLE":
                var_names.append(arg.value)
        if not var_names:
            var_names = [self.default_variable]
        return var_names

    def _extract_var(self, args):
        var_names = self._extract_vars(args)
        assert len(var_names) == 1
        return var_names[0]

    def _assert_and_extract_single(self, arg_type, args):
        items = [
            arg.value
            for arg in args
            if hasattr(arg, "type") and arg.type == self.token_prefix + arg_type
        ]
        assert len(items) <= 1
        return items.pop() if items else None

    def _extract_attribute(self, args):
        # extract a single attribute from the args
        return self._assert_and_extract_single("ATTRIBUTE", args)

    def _extract_entity_type(self, args):
        # extract a single entity type from the args
        return self._assert_and_extract_single("ENTITY_TYPE", args)

    def _extract_direction(self, args):
        # extract sort direction from args
        # default direction if no variable is found
        # return: if descending
        ds = [
            x
            for x in args
            if hasattr(x, "type")
            and (
                x.type == self.token_prefix + "ASC"
                or x.type == self.token_prefix + "DESC"
            )
        ]
        assert len(ds) <= 1
        d = ds.pop().type if ds else self.token_prefix + self.default_sort_order
        return True if d == self.token_prefix + "ASC" else False

    def _extract_if_reversed(self, args):
        rs = [
            x
            for x in args
            if hasattr(x, "type") and x.type == self.token_prefix + "REVERSED"
        ]
        return True if rs else False


def _first(args):
    return args[0].value


def _second(args):
    return args[1].value


def _third(args):
    return args[2].value


def _fourth(args):
    return args[3].value


def _fifth(args):
    return args[4].value


def _last(args):
    return args[-1].value


def _extract_entity_and_attribute(s):
    if ":" in s:
        etype, _, attr = s.partition(":")
    else:
        etype, attr = None, s
    return etype, attr
